#include <stdio.h>
#include <nbody.h>

const static int MAX_NUM_THREADS = 128;

#define safe(err)	__cudaSafeCall(err, __FILE__, __LINE__)
inline void __cudaSafeCall(cudaError_t err, const char * file, const int line) {
	if( cudaSuccess != err) {
        fprintf(stderr, "cudaSafeCall() Runtime API error in file <%s>, line %i : %s.\n",
                file, line, cudaGetErrorString( err) );
        exit(1);
    }
}

/* (inline device function) */
__device__ void calculate_force(
         Particle* this_particle1, Particle* this_particle2,
         float* force_x, float* force_y, float* force_z)
{ /* Particle_calculate_force */
    float difference_x, difference_y, difference_z;
    float distance_squared, distance;
    float force_magnitude;

    difference_x =
        this_particle2->position_x - this_particle1->position_x;
    difference_y =
        this_particle2->position_y - this_particle1->position_y;
    difference_z =
        this_particle2->position_z - this_particle1->position_z;

    distance_squared = difference_x * difference_x +
                       difference_y * difference_y +
                       difference_z * difference_z;

    distance = sqrtf(distance_squared);

    force_magnitude =
        gravitational_constant *
        (this_particle1->mass) * (this_particle2->mass) / distance_squared;

    *force_x = (force_magnitude / distance) * difference_x;
    *force_y = (force_magnitude / distance) * difference_y;
    *force_z = (force_magnitude / distance) * difference_z;

} /* Particle_calculate_force */

__global__ void calculate_forces_kernel_naive(Particle* d_particles, int number_of_particles) {
	size_t id = (blockDim.x * blockIdx.x) + threadIdx.x;
	
	if (id > number_of_particles) return;
	
	Particle* this_particle = d_particles + id;	
	
	float force_x = 0.0f, force_y = 0.0f, force_z = 0.0f;
	float total_force_x = 0.0f, total_force_y = 0.0f, total_force_z = 0.0f;
	
	for (int i = 0; i < number_of_particles; i++) {
		if (i != id) {
			calculate_force(this_particle, d_particles + i, &force_x, &force_y, &force_z);
			
			total_force_x += force_x;
			total_force_y += force_y;
			total_force_z += force_z;
		}
	}
	
	this_particle->force_x = total_force_x;
	this_particle->force_y = total_force_y;
	this_particle->force_z = total_force_z;
}

const size_t TILE_SIZE = 256; 	//number of particle positions to store in shared memory
								//note: must be a multiple of the number of threads per block

__global__ void calculate_forces_kernel(Particle* d_particles, int number_of_particles) {
	//__shared__ float4 tile[TILE_SIZE];
	//doing it this way eliminates warp serializing (doesn't speed it up at this point though)
	__shared__ float tile_x[TILE_SIZE];
	__shared__ float tile_y[TILE_SIZE];
	__shared__ float tile_z[TILE_SIZE];
	__shared__ float tile_w[TILE_SIZE];
	
	const size_t tile_per_thread = TILE_SIZE / blockDim.x; //(even multiple)
	
	size_t id = (blockDim.x * blockIdx.x) + threadIdx.x;
	
	//calculate the force for one particle from all other particles
	//Particle* this_particle = d_particles + id;
	
	// float me;
	// me_x = d_particles[id].position_x;
	// me_y = d_particles[id].position_y;
	// me_z = d_particles[id].position_z;
	// me_w = d_particles[id].mass;
	float 	me_x = d_particles[id].position_x,
			me_y = d_particles[id].position_y,
			me_z = d_particles[id].position_z,
			me_w = d_particles[id].mass;
	
	//float3 total = {0.0f, 0.0f, 0.0f};
	//float3 diff;
	 float total_x = 0.0f, total_y = 0.0f, total_z = 0.0f,
	 	  diff_x, diff_y, diff_z;
	float r2, rr, force_magnitude; // r2 == r^2, rr == 1/r (reciprocal of r)
	
	short t, s, i, tile_id, global_id;
	
	//tiling loop
	for (t = 0; t < number_of_particles; t += TILE_SIZE) {
		//copy values into tile
		for (s = 0; s < tile_per_thread; s++) {
			tile_id = threadIdx.x + s*blockDim.x;
			global_id = t + threadIdx.x + s*blockDim.x;
			tile_x[tile_id] = d_particles[global_id].position_x;
			tile_y[tile_id] = d_particles[global_id].position_y;
			tile_z[tile_id] = d_particles[global_id].position_z;
			tile_w[tile_id] = d_particles[global_id].mass;
		}
		__syncthreads(); //wait for everyone in block to finish loading
		
		if (id < number_of_particles) {
			for (i = 0; i < TILE_SIZE && (t + i) < number_of_particles; i++) {
				diff_x = tile_x[i] - me_x;
				diff_y = tile_y[i] - me_y;
				diff_z = tile_z[i] - me_z;
				r2 = diff_x*diff_x + diff_y*diff_y + diff_z*diff_z;
				rr = 1/sqrtf(r2);
				
				force_magnitude = gravitational_constant * tile_w[i] / r2;
				
				if (t+i != id) {
					total_x += force_magnitude * rr * diff_x;
					total_y += force_magnitude * rr * diff_y;
					total_z += force_magnitude * rr * diff_z;
				}
			}
		}
		__syncthreads();
	}
	if (id < number_of_particles) {
		d_particles[id].force_x = total_x * me_w;
		d_particles[id].force_y = total_y * me_w;
		d_particles[id].force_z = total_z * me_w;
	}
}

extern "C"
void Particle_array_calculate_forces_cuda(Particle* this_particle_array, int number_of_particles, int mpi_rank, int mpi_size, char* mpi_node_name) {
	/*	
	determine which GPU to run on:
	assume that each node will have the same number of GPUs available
	also assume that MPI processes are numbered node by node (for exam-
	ple, if each node has 2 GPUs on it, then it is assumed that we are 
	running 2 MPI processes per node, and that Node 0 has MPI processes
	ranked 0 & 1, Node 1 has ranks 2 & 3, etc.) */
	static bool deviceIsSet = false;
	
	int gpus_per_node;
	safe( cudaGetDeviceCount(&gpus_per_node) );
	int num_nodes = mpi_size / gpus_per_node;
	int gpu_index = mpi_rank / num_nodes;
	if (!deviceIsSet) {
		safe( cudaSetDevice(gpu_index) );
		deviceIsSet = true;
		printf("MPI %d (%s): gpus_per_node: %d; gpu_index: %d\n", mpi_rank, mpi_node_name, gpus_per_node, gpu_index);
	}
	
	const size_t SIZE = sizeof(Particle) * number_of_particles;
	size_t num_threads, num_blocks;
	
	//allocate device memory for particles
	Particle* d_particle_array;
	safe( cudaMalloc(&d_particle_array, SIZE) );

	//copy particle array over to GPU
	safe( cudaMemcpy(d_particle_array, this_particle_array, SIZE, cudaMemcpyHostToDevice) );
		
	//calculate block sizes
	if (number_of_particles < MAX_NUM_THREADS) {
		num_threads = number_of_particles;
		num_blocks = 1;
	} else {
		num_threads = MAX_NUM_THREADS;
		num_blocks = number_of_particles / MAX_NUM_THREADS 
						+ (number_of_particles % MAX_NUM_THREADS == 0 ? 0 : 1);
	}
	
	cudaEvent_t cstart, cend;
	float duration;
	cudaEventCreate(&cstart); cudaEventCreate(&cend);
	cudaEventRecord(cstart,0);
	
	//invoke kernel
	calculate_forces_kernel_naive <<< num_blocks, num_threads >>> (d_particle_array, number_of_particles);
	cudaThreadSynchronize();
	
	cudaEventRecord(cend,0);
	cudaEventSynchronize(cend);
	cudaEventElapsedTime(&duration, cstart, cend);
	cudaEventDestroy(cstart); cudaEventDestroy(cend);
	printf("kernel time (ms): %14g\n", duration);
	
	//get results back
	safe( cudaMemcpy(this_particle_array, d_particle_array, SIZE, cudaMemcpyDeviceToHost) );
	
	//free device memory
	cudaFree(d_particle_array);
	
}
